#include <iostream>
#include <cstdio>
#include <cstdlib>
#include <stdint.h>
#include <string>
#include <map>
#include <boost/optional.hpp>
#include <fstream>
#include <sys/time.h>
#include <time.h>

using namespace std;
using namespace boost;

void die(const string& reason)
{
  fprintf(stderr,"exiting because: %s", reason.c_str());
  exit(1);
}

struct InflightRequest
{
  InflightRequest() : count(0), pid(0)
  {
    startTime.tv_sec = endTime.tv_sec = 0;
    startTime.tv_usec = endTime.tv_usec = 0;
  }

  unsigned int count;
  struct timeval startTime, endTime;
  int pid;
};

typedef map<string, InflightRequest> inflight_t;
inflight_t inflight;

int main()
{
  char line[450], filename[450], device[20], sense[5], process[80];
  char pointer[20], fpointer[20];
  char mode[20];


  unsigned long stop, start;
  int bytes;
  int pid;
  int ret;

  optional<struct timeval> lowTime, lastTime;


  long totalBytesReadStarted=0, totalBytesReadEnded=0;
  long totalBytesWriteStarted=0, totalBytesWriteEnded=0;

  ofstream inflightplot("inflight.dat");
  ofstream headplaces("headplaces.dat");
  ofstream startplaces("startplaces.dat");
  ofstream latencies("latency.dat");
  double waitingTime=0;


  struct timeval tv;
  double nowMS=0;
  double msdiff;
  int nr_running, nr_iowait;
  int linenumber=0;
  string lastpointer;
  while(fgets(line, sizeof(line)-1, stdin) && *line) {
    linenumber++;
    if((ret=sscanf(line, "%s %ld %ld", mode, &tv.tv_sec, &tv.tv_usec))!=3) {
      printf("ret: %d, line number: %d\n", ret, linenumber);
      die("odd line: '"+string(line)+"'");
    }
    
    if(!lowTime)
      lowTime = tv;
    
    if(lastTime)
      msdiff = 1000*(tv.tv_sec - lastTime->tv_sec) + (tv.tv_usec - lastTime->tv_usec)/1000;

    nowMS = 1000*(tv.tv_sec - lowTime->tv_sec) + (tv.tv_usec - lowTime->tv_usec)/1000;

    if(!inflight.empty())
      waitingTime += msdiff;


    if(*mode == 'o') {
      if(sscanf(line, "%s %ld %ld %d %s", mode, &tv.tv_sec, &tv.tv_usec, &pid, filename)!=5)
	die("reading 'open' line: '"+string(line)+"'");
    }
    else if(*mode=='s') { //  start: 4294893535 1145781265 cefdb3c0 swapper(1): READ block 65 - 67 (1024) on hda
      if((ret=sscanf(line, "%s %ld %ld %s %s %s %s block %lu - %lu (%d) on %s %d %d", 
		     mode, &tv.tv_sec, &tv.tv_usec, pointer, fpointer, process, sense, &start, &stop, &bytes, device,
		     &nr_running, &nr_iowait
		))!=13) {
	printf("ret: %d\n", ret);
	die("reading 'start' line: '"+string(line)+"'");
      }
      lastpointer=pointer;
      if(!strcmp(sense,"READ"))
	totalBytesReadStarted+=bytes;
      else
	if(!strcmp(sense,"WRITE"))
	  totalBytesWriteStarted+=bytes;

      InflightRequest& ir=inflight[pointer];
      if(ir.count)
	cerr<<"Duplicate request on line "<<linenumber<<endl;

      ir.count++;
      ir.startTime=tv;

      ir.pid=atoi(strchr(process,'(')+1);

      if(*sense=='R')
	startplaces<< nowMS << "\t" << start<<"\n";
    }
    else if(*mode=='e') { // end: 4294893537 1153436309 cefdb3c0 READ 65 - 67 (1024) on hda
      if((ret=sscanf(line, "%s %ld %ld %s %s %lu - %lu (%d) on %s", 
		     mode, &tv.tv_sec, &tv.tv_usec, pointer, sense, &start, &stop, &bytes, device))!=9) {
	printf("ret: %d\n", ret);
	die("reading 'end' line: '"+string(line)+"'");
      }

      InflightRequest& ir=inflight[pointer];
      ir.endTime=tv;

      int val=--ir.count;
      if(val < 0)
	cerr<<"Saw request end that did not start on line "<<linenumber<<endl;


      if(pointer==lastpointer) {
	double latency = 1000 * (ir.endTime.tv_sec - ir.startTime.tv_sec) +
	  (ir.endTime.tv_usec - ir.startTime.tv_usec) / 1000.0;

	if(latency > 100)
	  cerr<<"Large latency "<<latency<<" on line "<<linenumber<<endl;
	latencies << latency<<"\n";
      }
      lastpointer=pointer;
      if(!val)
	inflight.erase(pointer);

      if(*sense=='R')
	totalBytesReadEnded+=bytes;
      else
	totalBytesWriteEnded+=bytes;

      if(lowTime)
	inflightplot << nowMS << "\t" << inflight.size()<<"\n";
      
      if(lowTime && *sense=='R')
	headplaces<< nowMS << "\t" << start<<"\n";



    }
    else
      die("unknown mode");


    lastTime = tv;

  }
  
  cerr<<"Total read started: "<<totalBytesReadStarted<<", total ended: "<< totalBytesReadEnded<<endl;
  cerr<<"Total write started: "<<totalBytesWriteStarted<<", total ended: "<< totalBytesWriteEnded<<endl;
  cerr<<"Waiting time: "<<waitingTime/1000.0<<" seconds"<<endl;
}
